/*
 * Jointrank -- Screening top fold changes using intersection and ranking
 * 
 * Copyright (C) 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>
 *
 *
 * Jointrank is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Jointrank is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Keywords: Gene Expression Experiments, Gene Expression Profile, BeadChip, BeadStudio
//
//
// jointrank.cpp
// last update: Wed Oct  6 18:17:07 IRST 2010

#define VERSION		1.2
#define COPYRIGHT	"Copyright 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>"

#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <cmath>
#include <string>
#include <vector>
#include <set>
#include <map>
#include <algorithm>
#include <cfloat>

using namespace std;

#define NVN				-1
#define isValidVal(x)	((x)>=0)

typedef char **argvType;
bool parseArgs ( int &argc, argvType & argv, const char *param ) {
    bool r = false;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = true;
            for ( j = i + 1; j < argc; ++j )
                argv[j - 1] = argv[j];
            --argc;
            --i;
        }
    }
    return r;
}

string parseArgsStr ( int &argc, argvType & argv, const char *param, const string &dfault ) {
    string r = dfault;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = argv[i + 1];
            for ( j = i + 2; j < argc; ++j )
                argv[j - 2] = argv[j];
            argc -= 2;
            i -= 2;
        }
    }
    return r;
}

int parseArgsInt ( int &argc, argvType & argv, const char *param, int dfault ) {
    int r = dfault;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = atoi ( argv[i + 1] );
            for ( j = i + 2; j < argc; ++j )
                argv[j - 2] = argv[j];
            argc -= 2;
            i -= 2;
        }
    }
    return r;
}

double parseArgsNum ( int &argc, argvType & argv, const char *param, double dfault ) {
    double r = dfault;
    int i, j;

    for ( i = 1; i < argc; ++i ) {
        for ( j = 0; argv[i][j] == param[j] && param[j] != '\0'; ++j );
        if ( param[j] == '\0' && argv[i][j] == '\0' ) {
            r = atof ( argv[i + 1] );
            for ( j = i + 2; j < argc; ++j )
                argv[j - 2] = argv[j];
            argc -= 2;
            i -= 2;
        }
    }
    return r;
}

double str2num(const string &t){
  double x=0;
  double p=0;
  string::const_iterator i;
  for ( i=t.begin(); i!=t.end(); i++ ){
	if(*i=='.' && p==0)
	  p=0.1;
	else if('0'<=*i && *i<='9'){
	  if(p>0){
		x+=p*(*i-'0');
		p/=10;
	  }
	  else
		x=10*x+(*i-'0');
	}
	else
	  return NVN;
  }
  return x;
}

vector<string> tokenize(string s, string delim){
  vector<string> v;
  string t="";
  string::iterator i;
  bool flag=false;
  for(i=s.begin();i!=s.end();i++){
    string::iterator j;
    for(j=delim.begin(); j!=delim.end() && *i!=*j; j++)
      ;
    if(flag && j!=delim.end()){
      v.push_back(t);
      t="";
    }
    flag=(j==delim.end());
    if(flag)
      t=t+*i;
  }
  if(flag)
    v.push_back(t);
  return v;
}

string sayPercent( int n, int N ) {
  ostringstream say;
  n*=100;
  if(0<n && n<N)
	say << "<1%";
  else
	say << n/N << "%";
  return say.str();
}

struct FoldChangeEnt{
  int index;
  double fc;
  double efc;
  bool operator<(const FoldChangeEnt &x) const {
    return efc<x.efc;
  }
};

struct rec{
  map<string,string> noteList;
  vector<FoldChangeEnt> fcList;
  static string noteRep;
  static int fcRep;
  bool operator<(const rec &x) const{
    int i;
    for(i=fcRep; i>=0 && fabs(fcList[i].efc-x.fcList[i].efc)<0.001; i--)
      ;
    return fcList[i].efc>x.fcList[i].efc;
  }
  bool isConsistFC( double threshold, int copred ){
    vector<FoldChangeEnt>::iterator i;
    for(i=fcList.begin();i!=fcList.end();i++)
      if(1.0/threshold < i->fc && i->fc < threshold)
	copred--;
    if(copred<0) return false;
    for(i=fcList.begin();i!=fcList.end() && i->fc==1.0;i++)
      ;
    if(i!=fcList.end()){
      double r = log(i->fc);
      for( ; i!=fcList.end() && r*log(i->fc)>=0;i++)
		;
    }
	return i==fcList.end();
  }
  void cutoff( double threshold ){
    vector<FoldChangeEnt>::iterator j;
    for(j=fcList.begin();j!=fcList.end();j++)
      if(j->efc<threshold) j->efc=1.0;
  }
  void fillin(int n){
	vector<bool> f(n);
    vector<FoldChangeEnt>::iterator j;
    for(j=fcList.begin();j!=fcList.end();j++)
	  f[j->index]=true;
	FoldChangeEnt o;
	o.fc=o.efc=1.0;
	int index;
	for(index=0; index<n; index++)
	  if(!f[index]){
		o.index=index;
		fcList.push_back(o);
	  }
  }
  void inverse(){
    vector<FoldChangeEnt>::iterator j;
    for(j=fcList.begin();j!=fcList.end();j++)
      j->efc=1.0/j->fc;
  }
};

string rec::noteRep;
int rec::fcRep;

bool symbolCmp( const rec& a, const rec &b ) {
  return a.noteList.find(rec::noteRep)->second < b.noteList.find(rec::noteRep)->second;
}


class logstream : public stringstream {
public:
  string label;
  ostream *os;
};

ostream &log( ostream &o ){
  logstream *l = (logstream *) (&o);
  if(l->str()!="")
    *(l->os) << l->label;
  *(l->os) << l->str() << endl;
  l->str("");
  return *(l->os);
}

class Jointrank{
  double thrsh;
  int top;
  int pred, copred;
  bool fancy;
  string arr;
  string FC;
  string preFCLabel;

  ifstream *ifs;
  vector<string> infiles;
  int nin, curfile, l;
  
  ostream *ofs;
  string outfile;  
  string delim;

  map<string,int> notes;
  vector<rec> probe;
  vector<string> fcEnt;
  vector<int> numenator;
  vector<int> denominator;
  int nField, nVal, nVal0;
  int passed;

  set<string> keywords;
  vector<string> h;
  vector<string> t;
  vector<double> v;
  int i0,i1;

  int maxIndex;
  bool crossFile;

  logstream Jlog;
  bool headPrintedOnce;

public:
  enum regDir{down,up};

  void printUsage();
  bool setup( int &argc, argvType & argv );
  bool tryLine();
  void setupRules();
  void readLine();
  bool readFile();
  void detectSymbolField();
  void purge();
  void fillinFC();
  void inverseFC();
  void analysis();
  void printSum(const regDir dir);
  void printHead();
  void print();
};

void Jointrank::printUsage(){
    cerr << "Jointrank " << VERSION << endl;
    cerr << COPYRIGHT << endl;
    cerr << endl;
    cerr << "  Usage:" << endl;
    cerr << "        jointrank [options] -FC <list>" << endl;
    cerr << endl;
    cerr << "  Options:" << endl;
    cerr << "        -i or -infile <file>     The input file of expression data." << endl;
    cerr << "                                 If your data are provided by several files," << endl;
    cerr << "                                 give a comma separated list of them." << endl;
    cerr << "                                 (default: input by keyboard)" << endl;
    cerr << "        -delim \"xyz...\"          Field delimiters. Each character of \"xyz...\"" << endl;
    cerr << "                                 separates fields in input file. (default: tab)" << endl;
    cerr << "        -o or -outfile <file>    Output file. (default: write to display screen)" << endl;
    cerr << "                                 A result summary is also reported by another" << endl;
    cerr << "                                 output with the same file name and \".summary\"" << endl;
    cerr << "                                 extension." << endl;
    cerr << "        -label <a>               The Fold Change columns in the output file are" << endl;
	cerr << "                                 labeled by titles whose prefix is <a>." << endl;
	cerr << "                                 (default: -label F)." << endl;
    cerr << "        -a or -arr <key>         Arrange the output list based on key." << endl;
    cerr << "                                 The key can be one of or a combination of two" << endl;
    cerr << "                                 letters \'s\' or \'f\'. By \'s\', selected genes" << endl;
    cerr << "                                 are alphabetically sorted based on gene" << endl;
    cerr << "                                 symbols and by \'f\', they are sorted based on" << endl;
    cerr << "                                 their Fold Changes." << endl;
    cerr << "                                 Note that the order of these letters in a" << endl;
    cerr << "                                 combined key is significant such that -a sf" << endl;
    cerr << "                                 results a different arrangement than -a fs." << endl;
    cerr << "                                 (default: -a f)." << endl;
    cerr << "        -fancy                   Put some fancy figures into output." << endl;    
    cerr << endl;
    cerr << "        -thrsh <threshold>       Fold change threshold." << endl;
    cerr << "                                 (default: 1.0; threshold free fold changes)" << endl;
    cerr << "        -top <n>                 Select n top fold changes." << endl;
    cerr << "                                 (default: 1000; a trivial top list!)" << endl;
    cerr << endl;
    cerr << "        -pred <k>                The selection will be predicated " << endl;
    cerr << "                                 on k observations." << endl;
    cerr << "                                 Here, k is an integer or \"all\" (default)." <<endl;
    cerr << endl;
    cerr << "  <list> a comma separated list of column indexes and/or column headers on" << endl;
    cerr << "         which the program evaluates fold changes and sorts affected genes" << endl;
    cerr << "         based." << endl;
    cerr << "         The list entries are in the following format:" << endl;
    cerr << endl;
    cerr << "           TreatId/ControlId     " << endl;
    cerr << endl;
    cerr << "         Each entry of the list simply represents a formula for a fold change." << endl;
    cerr << "         You should replace TreatId and ControlId with integers denoting" << endl;
    cerr << "         indexes of two columns respectively corresponded to a treatment sample" << endl;
    cerr << "         and a control sample." << endl;
    cerr << "         Conveniently, you can use column headers instead of column indexes." << endl;
    cerr << endl;
    cerr << "  If your input file contains pre-calculated fold changes instead of absolute" << endl;
    cerr << "  signal values, simply ignore \"/ControlId\" and make a list of single column ids." << endl;
    cerr << endl;
    cerr << "  You don't have to use all columns of the input file for your assessment. You" << endl;
    cerr << "  can direct several assessments based on partial subsets of treatments and " << endl;
    cerr << "  controls using a single data file." << endl;
    cerr << endl;
    cerr << "  The program properly behaves with generic formats of gene expression data." << endl;
    cerr << "  For instance, it is no matter whether a header row is distinguished in the" << endl;
    cerr << "  input file or not." << endl;
}

bool Jointrank::setup(int &argc, argvType & argv) {
  FC = parseArgsStr( argc, argv, "-FC", "" );
  FC = parseArgsStr( argc, argv, "-fc", FC );
  if( argc==1 || parseArgs( argc, argv, "-h" ) || parseArgs( argc, argv, "--help" ) || FC.empty() ){
    printUsage();
    return false;
  }
  thrsh = parseArgsNum( argc, argv, "-thrsh", 1.0 );
  top = parseArgsInt( argc, argv, "-top", 1000 );
  pred = parseArgsInt ( argc, argv, "-pred", 0 );
  fancy = parseArgs( argc, argv, "-fancy" );
  arr = parseArgsStr( argc, argv, "-a", "f" );
  arr = parseArgsStr( argc, argv, "-arr", arr );

  preFCLabel = parseArgsStr( argc, argv, "-label", "F" );

  string infile = parseArgsStr( argc, argv, "-i", "-" );
  infile = parseArgsStr( argc, argv, "-infile", infile );
  if(infile=="-")
    ifs=(ifstream *)(&cin);
  else{
	infiles = tokenize(infile,",");
	nin=infiles.size();
	ifs = new ifstream[nin];
	int i;
	for(i=0;i<nin;i++){
	  ifs[i].open(infiles[i].c_str());
	  if(!ifs[i]){
		cerr << "Cannot open " << infiles[i] << endl;
		return false;
	  }
	}
  }
  
  outfile = parseArgsStr( argc, argv, "-o", "-" );
  outfile = parseArgsStr( argc, argv, "-outfile", outfile );
  if(outfile=="-"){
    Jlog.label="### ";
    ofs=Jlog.os=&cout;
  }
  else{
    Jlog.label="";
    ofs = new ofstream(outfile.c_str());
    if(!(*ofs)){
      cerr << "Cannot open " << outfile << endl;
      return false;
    }
    string summaryFile = outfile + ".summary";
    Jlog.os = new ofstream(summaryFile.c_str());
    if(!(*(Jlog.os))){
      cerr << "Cannot open " << summaryFile << endl;
      return false;
    }
  }

  delim = parseArgsStr( argc, argv, "-delim", "" );
  delim = parseArgsStr( argc, argv, "-delimit", delim );

  Jlog << "Options of the joint-rank analysis are as follows;" << log;
  Jlog << "-FC " << FC << log;
  Jlog << "-i " << infile << log;
  Jlog << "-o " << outfile << ' ';
  Jlog << "-label " << preFCLabel << log;
  Jlog << "-a " << arr << ' ';
  if(fancy)  Jlog << "-fancy " << ' ';
  if(pred==0)
    Jlog << "-pred all" << ' ';
  else
  Jlog << "-pred " << pred << ' ';
  Jlog << "-top " << top <<  ' ';
  Jlog << "-thrsh " << thrsh <<  ' ';
  Jlog << "-delim \"" << delim << "\t\"" << log;
  Jlog << log;

  delim = delim + "\'\"\r\t\n";

  fcEnt = tokenize(FC,",");
  vector<string>::iterator i;
  for(i=fcEnt.begin();i!=fcEnt.end();i++){
    vector<string> fcIns = tokenize(*i,"\'\"/");
    int a,b;
    a=str2num(fcIns[0]);
    if(fcIns.size()<2)
      b=NVN;
    else
      b=str2num(fcIns[1]);
    if(isValidVal(a)){
	  numenator.push_back(a-1);
      denominator.push_back(b-1);
    }
  }

  if(pred==0) pred=fcEnt.size();
  copred=fcEnt.size()-pred;
  rec::fcRep=copred;

  headPrintedOnce=false;
  return true;
}

bool Jointrank::tryLine(){
  t.clear();
  v.clear();

  string s;
  if(getline(*ifs,s)==false)
    return false;
  
  t = tokenize(s,delim);
  if(t.size()>0){
	if(nField==0) nField=t.size();
	if(t.size()!=nField){
	  cerr << "Error: the number of fields in row " << l << " file " << infiles[curfile] << " is inconsistent with previous rows." << endl;
	  return false;
	}

	nVal0=0;
	vector<string>::iterator i;
	for(i=t.begin();i!=t.end();i++){
	  double x=str2num(*i);
	  v.push_back(x);
	  if(isValidVal(x))
		nVal0++;
	}
	if(nVal0>0){
	  if(nVal==0) nVal=nVal0;
	  if(nVal0!=nVal){
		cerr << "Error: the number of numeric fields in row "  << l << " file " << infiles[curfile] << " is inconsistent with previous rows." << endl;
		return false;
	  }
	}
  }

  return true;
}

int parse(const string &f, const vector<string> &t){
  vector<string>::const_iterator i=find(t.begin(),t.end(),f);
  if(i==t.end())
	return NVN;
  else if(find(i+1,t.end(),f)!=t.end()){
	cerr << "Redundent column header: " << f << endl;
	return NVN;
  }
  else
	return i-t.begin();
}

void Jointrank::setupRules() {
  i0=h.size();
  vector<string>::iterator i;
  for(i=t.begin();i!=t.end();++i)
	h.push_back(*i);
  i1=h.size();
  for(i=fcEnt.begin();i!=fcEnt.end();i++){
    vector<string> fcIns = tokenize(*i,"\'\"/");
    int a,b;
	a=parse(fcIns[0],h);
	if(isValidVal(a) && i0<=a && a<i1){
	  if(fcIns.size()<2)
		b=NVN;
	  else
		b=parse(fcIns[1],h);
	  if(!isValidVal(b) || i0<=b && b<i1){
		numenator.push_back(a);
		denominator.push_back(b);
	  }
	}
  }
}

void Jointrank::readLine(){
  int i,j;
  if(i0==i1){
	i1=i0+nField;
	for(i=i0;i<i1;i++){
	  ostringstream keyword;
	  keyword << "keyword" << i;
	  h.push_back(keyword.str());
	}
  }  

  rec rc;
  string s;
  for(j=0;j<v.size();j++){
    if(!isValidVal(v[j])){
	  s+=t[j];
	  rc.noteList[*(keywords.insert(h[j+i0]).first)]=t[j];
	}
  }
  if(notes.find(s)==notes.end()){
	notes[s]=probe.size();
	probe.push_back(rc);
  }
  i=notes[s];
  

  for(j=0;j<numenator.size();j++){
	int a=numenator[j];
    if(maxIndex<a)
	  maxIndex=a;
	int b=denominator[j];
	if(!isValidVal(b)){
	  a-=i0; 
	  if(0<=a && a<nField){
		FoldChangeEnt z;
		z.index=j;
		z.fc=z.efc=v[a];
		probe[i].fcList.push_back(z);
	  }
	}
	else{
      if(maxIndex<b)
	    maxIndex=b;
  	  a-=i0;
	  b-=i0;
	  if((0<=a && a<nField) && (0<=b && b<nField)){
		FoldChangeEnt z;
		z.index=j;
		z.fc=z.efc=v[a]/v[b];
		probe[i].fcList.push_back(z);
	  }
	  else if((0<=a && a<nField) || (0<=b && b<nField))
		crossFile=true;		
	}
  }
}

bool Jointrank::readFile(){
  maxIndex=0;
  crossFile=false;
  i0=0;
  int n=0;
  for(curfile=0;curfile<nin;ifs++,curfile++){
	nField=nVal=l=0;
	i1=i0;
	while(tryLine()){
	  if(t.size()==0) continue;
	  if(nVal0==0)
		setupRules();
	  else{
		readLine();
		l++;
		n++;
	  }
	}
	if(!ifs->eof()){
	  cerr << "Error while reading file " << infiles[curfile] << endl;
	  return false;
	}
  }
  if(curfile<nin){
	cerr << "Error while reading file " << infiles[curfile] << endl;
	return false;
  }

  if(fcEnt.size()!=numenator.size()){
    cerr << "Error in parsing Fold Change formulae or misspelt column identifier." << endl;
    return false;
  }

  if(maxIndex>=h.size()){
    cerr << "Error in Fold Change formula: index out of range." << endl;
    return false;
  }

  if(crossFile){
    cerr << "Error: crossing file Fold Change evaluation is invalid." << endl;
    return false;
  }

  // First purge inconsistent fold change entries 
  purge();
  // Then fill-in un-assigned entries with 1.0 
  fillinFC();

  if(top>probe.size()) top=probe.size();

  Jlog << "Defined " << fcEnt.size() << " Fold Change formulae:" << log;
  int i;
  for(i=0;i<fcEnt.size();i++){
    Jlog << preFCLabel << (i+1) << ":\t" << h[numenator[i]];
	if(isValidVal(denominator[i]))
	  Jlog << '/' << h[denominator[i]];
	Jlog << log;
  }

  Jlog << n << " probes read; " << probe.size() << " probes accepted (" << sayPercent(probe.size(),n) << ")" << log;
  Jlog << log;

  return true;
}

void Jointrank::detectSymbolField(){
  set<string> s;
  set<string>::iterator i,k;
  int nmin=probe.size();
  vector<rec>::iterator r;
  for(i=keywords.begin();i!=keywords.end();i++){
    s.clear();
    for(r=probe.begin();r!=probe.begin();r++)
      s.insert(r->noteList[*i]);
    if(nmin>s.size()){
      nmin=s.size();
      k=i;
    }
  }
  rec::noteRep=*k;  
}

void Jointrank::purge() {
  int i,j;
  j=probe.size();
  for(i=0;i<j;i++){
	if( ! probe[i].isConsistFC(thrsh,copred) ){
	  j--;
	  probe[i]=probe[j];
	  i--;
	}
  }
  probe.resize(j);
}

void Jointrank::fillinFC() {
  vector<rec>::iterator r;
  for(r=probe.begin();r!=probe.end();r++)
    r->fillin(fcEnt.size());
}

void Jointrank::inverseFC(){
  vector<rec>::iterator r;
  for(r=probe.begin();r!=probe.end();r++)
    r->inverse();
}

void Jointrank::analysis(){
  vector<rec>::iterator r;
  for(r=probe.begin();r!=probe.end();r++)
    r->cutoff(thrsh);
  for(r=probe.begin();r!=probe.end();r++)
    sort(r->fcList.begin(),r->fcList.end());

  partial_sort(probe.begin(),probe.begin()+top,probe.end());  

  vector<FoldChangeEnt>::iterator j;
  for(r=probe.begin();r!=probe.begin()+top;r++){
	int n=0;
	for(j=r->fcList.begin();j!=r->fcList.end();j++)
	  if(j->efc>1.0)
		n++;
	  if(n<pred)
		break;
  }
  passed=r-probe.begin();
}

void Jointrank::printSum( const regDir dir ){
  if(dir==up) Jlog << "Up regulate genes:" << log;
  else        Jlog << "Down regulate genes:" << log;

  if(numenator.size()==1)
    Jlog << "Result genes are satisfied the following criterion;" << log;
  else if(pred==1)
    Jlog << "Result genes are satisfied at least one of the following criteria;" << log;
  else if(copred==0)
    Jlog << "Result genes are satisfied all the following criteria;" << log;
  else
    Jlog << "Result genes are satisfied at least " << pred << " of the following criteria;" << log;
  
  int i;
  for(i=0;i<numenator.size();i++){
    Jlog << i+1 << ")   " << preFCLabel << i+1;
    if(dir==up)
      Jlog << " >= " << thrsh;
    else // down
      Jlog << " <= " << 1.0/thrsh;
    Jlog << " and rank(" << preFCLabel << i+1 << ") <= " << top << log;
  }
  Jlog << log;

  vector<rec>::iterator r;
  vector<int> sum(numenator.size());
  for(r=probe.begin();r!=probe.end();r++){
    vector<FoldChangeEnt>::iterator j;
    for(j=r->fcList.begin();j!=r->fcList.end();j++)
      if(j->efc>1.0)
	sum[j->index]++;
  }
  if(passed==0)
      Jlog << "No probe found in this case." << log;
  else{
      Jlog << "Final result consists of " << passed << " probes that cover " << log;
       vector<int>::iterator it;
       for(it=sum.begin();it!=sum.end();it++){
	  //--- to do: better report about "not satisfied" but predicated genes  
          Jlog << sayPercent(passed,max(passed,*it)) << " of " << (*it) << " probes satisfying the criterion on " << preFCLabel << it-sum.begin()+1;
          if(sum.end()-it>2)
	    Jlog << "," << log;
          else if(sum.end()-it==2)
	    Jlog << ", and" << log;
          else 
	    Jlog << "." << log;
       }
  }
  Jlog << log;
}

void Jointrank::printHead(){
  if(!headPrintedOnce){
	set<string>::iterator it;
	(*ofs) << rec::noteRep << "\t";  
	for(it=keywords.begin(); it!=keywords.end(); it++)
	  if(*it!=rec::noteRep)
		(*ofs) << *it << "\t";
	  
	int i;
	for(i=0; i< numenator.size(); i++){
	  if(isValidVal(denominator[i]))
		(*ofs) << preFCLabel << (i+1) << '\t';
	  else
		(*ofs) << h[numenator[i]] << '\t';
	}
	
	if(fancy)
	  (*ofs) << "FC-Meter" << '\t';

	(*ofs) << endl;
	
	headPrintedOnce=(ofs!=Jlog.os);
  }
}

void Jointrank::print(){ 
  vector<rec>::iterator r;
  if(*(arr.rbegin())=='s')
	sort(probe.begin(),probe.begin()+passed,symbolCmp);
  if(arr.size()==2){
	if(arr[0]=='f'){
	  vector<rec>::iterator it=probe.begin();
	  double efcmax;
	  string sym="";
	  for(r=probe.begin();r!=probe.begin()+passed;r++){
		if(r->noteList.find(rec::noteRep)->second==sym)
		  efcmax = efcmax > r->fcList[rec::fcRep].efc ? efcmax:r->fcList[rec::fcRep].efc;
		else{
		  while(it!=r){
			it->fcList[rec::fcRep].efc=efcmax;
			it++;
		  }
		  sym=it->noteList.find(rec::noteRep)->second;
		  efcmax=it->fcList[rec::fcRep].efc;
		}
	  }
	  while(it!=r){
		it->fcList[rec::fcRep].efc=efcmax;
		it++;
	  }
	  stable_sort(probe.begin(),probe.begin()+passed);
	}
	if(arr[0]=='s')
	  stable_sort(probe.begin(),probe.begin()+passed,symbolCmp);    
  }
  for(r=probe.begin();r!=probe.begin()+passed;r++){
	set<string>::iterator i;
	(*ofs) << r->noteList.find(rec::noteRep)->second << "\t";  
	for(i=keywords.begin(); i!=keywords.end(); i++)
	  if(*i!=rec::noteRep)
		(*ofs) << r->noteList[*i] << "\t";  
	  
	vector<double> slot(numenator.size(),1.0);
	vector<FoldChangeEnt>::iterator j;
	double minv=DBL_MAX;
	double maxv=DBL_MIN;
	for(j=r->fcList.begin();j!=r->fcList.end();j++){
	  slot[j->index]=j->fc;
	  if(j->fc!=1.0){
		if(minv > j->fc)  minv=j->fc;
		if(maxv < j->fc)  maxv=j->fc;
	  }
	}
	vector<double>::iterator k;
	for(k=slot.begin();k!=slot.end();k++)
	  (*ofs) << *k << "\t";
	
	if(fancy){
	  double repv=r->fcList[rec::fcRep].fc;
	  if(repv < 1.0){
		repv=1.0/repv;
		double temp=minv;
		minv=1.0/maxv;
		maxv=1.0/temp;
	  }
	  (*ofs) << '-';
	  if(repv>1.0)
		(*ofs) << string(10*(minv-1.0),'=') << string(10*(repv-minv),'I') << '+' << string(10*(maxv-repv),'I');
	}
	(*ofs) << endl;
  }
  (*ofs) << endl;
}

int main(int argc, char *argv[]){
  Jointrank J;
  if(J.setup(argc, argv)==false)
    return 1;
  if(J.readFile()==false)
    return 1;

  J.detectSymbolField();
  
  J.analysis();
  J.printSum(Jointrank::up);
  J.printHead();
  J.print();

  J.inverseFC();

  J.analysis();
  J.printSum(Jointrank::down);
  J.printHead();
  J.print();
  
  return 0;
}
